"""View Module."""


from jinja2 import ChoiceLoader, Environment, PackageLoader, select_autoescape
from jinja2.exceptions import TemplateNotFound

from masonite.exceptions import RequiredContainerBindingNotFound


class View:
    """View class. Responsible for handling everything involved with views and view environments."""

    _splice = '/'

    def __init__(self, container):
        """View constructor.

        Arguments:
            container {masonite.app.App} -- Container object.
        """
        self.dictionary = {}
        self.composers = {}
        self.container = container

        # If the cache_for method is declared
        self.cache = False
        # Cache time of cache_for
        self.cache_time = None
        # Cache type of cache_for
        self.cache_type = None

        self.template = None
        self.environments = []
        self.extension = '.html'
        self._jinja_extensions = ['jinja2.ext.loopcontrols']
        self._filters = {}
        self._tests = {}

    def render(self, template, dictionary={}):
        """Get the string contents of the view.

        Arguments:
            template {string} -- Name of the template you want to render.

        Keyword Arguments:
            dictionary {dict} -- Data that you want to pass into your view. (default: {{}})

        Returns:
            self
        """
        self.__load_environment(template)

        self.dictionary.update(dictionary)

        # Check if use cache and return template from cache if exists
        if self.container.has('Cache') and self.__cached_template_exists() and not self.__is_expired_cache():
            return self.__get_cached_template()

        # Check if composers are even set for a speed improvement
        if self.composers:
            self._update_from_composers()

        if self._tests:
            self.env.tests.update(self._tests)

        self.rendered_template = self._render()

        return self

    def _render(self):
        try:
            # Try rendering the template with '.html' appended
            return self.env.get_template(self.filename).render(
                self.dictionary)
        except TemplateNotFound:
            # Try rendering the direct template the user has supplied
            return self.env.get_template(self.template).render(
                self.dictionary)

    def _update_from_composers(self):
        """Add data into the view from specified composers."""
        # Check if the template is directly specified in the composer
        if self.template in self.composers:
            self.dictionary.update(self.composers[self.template])

        # Check if there is just an astericks in the composer
        if '*' in self.composers:
            self.dictionary.update(self.composers['*'])

        # We will append onto this string for an easier way to search through wildcard routes
        compiled_string = ''

        # Check for wildcard view composers
        for template in self.template.split(self._splice):
            # Append the template onto the compiled_string
            compiled_string += template
            if '{}*'.format(compiled_string) in self.composers:
                self.dictionary.update(
                    self.composers['{}*'.format(compiled_string)])
            else:
                # Add a slash to symbolize going into a deeper directory structure
                compiled_string += '/'

    def composer(self, composer_name, dictionary):
        """Update composer dictionary.

        Arguments:
            composer_name {string} -- Key to bind dictionary of data to.
            dictionary {dict} -- Dictionary of data to add to controller.

        Returns:
            self
        """
        if isinstance(composer_name, str):
            self.composers[composer_name] = dictionary

        if isinstance(composer_name, list):
            for composer in composer_name:
                self.composers[composer] = dictionary

        return self

    def share(self, dictionary):
        """Share data to all templates.

        Arguments:
            dictionary {dict} -- Dictionary of key value pairs to add to all views.

        Returns:
            self
        """
        self.dictionary.update(dictionary)
        return self

    def cache_for(self, time=None, type=None):
        """Set time and type for cache.

        Keyword Arguments:
            time {string} -- Time to cache template for (default: {None})
            type {string} -- Type of the cache. (default: {None})

        Raises:
            RequiredContainerBindingNotFound -- Thrown when the Cache key binding is not found in the container.

        Returns:
            self
        """
        if not self.container.has('Cache'):
            raise RequiredContainerBindingNotFound(
                "The 'Cache' container binding is required to use this method and wasn't found in the container. You may be missing a Service Provider"
            )

        self.cache = True
        self.cache_time = float(time)
        self.cache_type = type
        if self.__is_expired_cache():
            self.__create_cache_template(self.template)
        return self

    def exists(self, template):
        """Check if a template exists.

        Arguments:
            template {string} -- Name of the template to check for.

        Returns:
            bool
        """
        self.__load_environment(template)

        try:
            self.env.get_template(self.filename)
            return True
        except TemplateNotFound:
            return False

    def add_environment(self, template_location, loader=PackageLoader):
        """Add an environment to the templates.

        Arguments:
            template_location {string} -- Directory location to attach the environment to.

        Keyword Arguments:
            loader {jinja2.Loader} -- Type of Jinja2 loader to use. (default: {jinja2.PackageLoader})
        """        # loader(package_name, location)
        # /dashboard/templates/dashboard
        if loader == PackageLoader:
            template_location = template_location.split(self._splice)

            self.environments.append(
                loader(template_location[0], '/'.join(template_location[1:])))
        else:
            self.environments.append(
                loader(template_location))

    def filter(self, name, function):
        """Use to add filters to views.

        Arguments:
            name {string} -- Key to bind the filter to.
            function {object} -- Function used for the template filter.
        """
        self._filters.update({name: function})

    def test(self, key, obj):
        self._tests.update({key: obj})
        return self

    def add_extension(self, extension):
        self._jinja_extensions.append(extension)
        return self

    def __load_environment(self, template):
        """Private method for loading all the environments.

        Arguments:
            template {string} -- Template to load environment from.
        """
        self.template = template
        self.filename = template.replace(self._splice, '/') + self.extension

        if template.startswith('/'):
            # Filter blanks strings from the split
            location = list(filter(None, template.split('/')))
            self.filename = location[-1] + self.extension

            loader = ChoiceLoader(
                [PackageLoader(location[0], '/'.join(location[1:-1]))] + self.environments
            )

            self.env = Environment(
                loader=loader,
                autoescape=select_autoescape(['html', 'xml']),
                extensions=self._jinja_extensions
            )

        else:
            loader = ChoiceLoader(
                [PackageLoader('resources', 'templates')]
                + self.environments
            )

            # Set the searchpath since some packages look for this object
            # This is sort of a hack for now
            loader.searchpath = ''

            self.env = Environment(
                loader=loader,
                autoescape=select_autoescape(['html', 'xml']),
                extensions=self._jinja_extensions
            )

        self.env.filters.update(self._filters)

    def __create_cache_template(self, template):
        """Save in the cache the template.

        Arguments:
            template {string} -- Creates the cached templates.
        """
        self.container.make('Cache').store_for(
            template, self.rendered_template,
            self.cache_time, self.cache_type, '.html',
        )

    def __cached_template_exists(self):
        """Check if the cache template exists.

        Returns:
            bool
        """
        return self.container.make('Cache').exists(self.template)

    def __is_expired_cache(self):
        """Check if cache is expired.

        Returns:
            bool
        """
        # Check if cache_for is set and configurate
        if self.cache_time is None or self.cache_type is None and self.cache:
            return True

        driver_cache = self.container.make('Cache')

        # True is expired
        return not driver_cache.is_valid(self.template)

    def __get_cached_template(self):
        """Return the cached version of the template.

        Returns:
            self
        """
        driver_cache = self.container.make('Cache')
        self.rendered_template = driver_cache.get(self.template)
        return self

    def set_splice(self, splice):
        self._splice = splice
        return self



"""The base class that all drivers inherit from."""


class BaseDriver:
    """Base driver class."""

    _manager = None

    def driver(self, driver):
        """Return an instance of the driver specified.

        Arguments:
            driver {string} -- String representation of the driver to be resolved from the container.
                               This can be values like "s3" or "disk"

        Returns:
            masonite.drivers -- Returns an instance of the driver.
        """
        return self._manager.driver(driver)

    def load_manager(self, manager):
        """Load the manager into the driver.

        Arguments:
            manager {masonite.managers} -- Needs to be a Manager class.

        Returns:
            self
        """
        self._manager = manager
        return self



from .AuthCommand import AuthCommand
from .BaseScaffoldCommand import BaseScaffoldCommand
from .CommandCommand import CommandCommand
from .ControllerCommand import ControllerCommand
from .DownCommand import DownCommand
from .InfoCommand import InfoCommand
from .InstallCommand import InstallCommand
from .JobCommand import JobCommand
from .KeyCommand import KeyCommand
from .MakeMigrationCommand import MakeMigrationCommand
from .MiddlewareCommand import MiddlewareCommand
from .MigrateCommand import MigrateCommand
from .MigrateRefreshCommand import MigrateRefreshCommand
from .MigrateResetCommand import MigrateResetCommand
from .MigrateRollbackCommand import MigrateRollbackCommand
from .MigrateStatusCommand import MigrateStatusCommand
from .ModelCommand import ModelCommand
from .ModelDocstringCommand import ModelDocstringCommand
from .ProviderCommand import ProviderCommand
from .QueueWorkCommand import QueueWorkCommand
from .ServeCommand import ServeCommand
from .ViewCommand import ViewCommand
from .ValidatorCommand import ValidatorCommand
from .RoutesCommand import RoutesCommand
from .SeedCommand import SeedCommand
from .SeedRunCommand import SeedRunCommand
from .TinkerCommand import TinkerCommand
from .UpCommand import UpCommand



"""New Authentication System Command."""
import os
import shutil

from cleo import Command


class AuthCommand(Command):
    """
    Creates an authentication system.

    auth
    """

    def handle(self):
        self.info('Scaffolding Application ...')
        module_path = os.path.dirname(os.path.realpath(__file__))

        f = open('routes/web.py', 'a')
        # add all the routes
        f.write('\nROUTES = ROUTES + [\n    ')
        f.write("Get().route('/login', 'LoginController@show'),\n    ")
        f.write("Get().route('/logout', 'LoginController@logout'),\n    ")
        f.write("Post().route('/login', 'LoginController@store'),\n    ")
        f.write("Get().route('/register', 'RegisterController@show'),\n    ")
        f.write("Post().route('/register', 'RegisterController@store'),\n    ")
        f.write("Get().route('/home', 'HomeController@show'),\n    ")
        f.write("Get().route('/email/verify', 'ConfirmController@verify_show'),\n    ")
        f.write("Get().route('/email/verify/@id:signed', 'ConfirmController@confirm_email'),\n    ")
        f.write("Get().route('/email/verify/@id:signed', 'ConfirmController@confirm_email'),\n    ")
        f.write("Get().route('/password', 'PasswordController@forget').name('forgot.password'),\n    ")
        f.write("Post().route('/password', 'PasswordController@send'),\n    ")
        f.write("Get().route('/password/@token/reset', 'PasswordController@reset'),\n    ")
        f.write("Post().route('/password/@token/reset', 'PasswordController@update'),\n")
        f.write(']\n')

        # move controllers
        shutil.copyfile(module_path + "/../snippets/auth/controllers/LoginController.py",
                        os.getcwd() + "/app/http/controllers/LoginController.py")
        shutil.copyfile(module_path + "/../snippets/auth/controllers/RegisterController.py",
                        os.getcwd() + "/app/http/controllers/RegisterController.py")
        shutil.copyfile(module_path + "/../snippets/auth/controllers/HomeController.py",
                        os.getcwd() + "/app/http/controllers/HomeController.py")
        shutil.copyfile(module_path + "/../snippets/auth/controllers/ConfirmController.py",
                        os.getcwd() + "/app/http/controllers/ConfirmController.py")
        shutil.copyfile(module_path + "/../snippets/auth/controllers/PasswordController.py",
                        os.getcwd() + "/app/http/controllers/PasswordController.py")
        # move templates
        shutil.copytree(module_path + "/../snippets/auth/templates/auth",
                        os.getcwd() + "/resources/templates/auth")

        self.info('Project Scaffolded. You now have 5 new controllers, 7 new templates and 9 new routes')



"""Upload Disk Driver."""

import os
import _io

from masonite.contracts import UploadContract
from masonite.drivers import BaseUploadDriver
from masonite.app import App


class UploadDiskDriver(BaseUploadDriver, UploadContract):
    """Upload to and from the file system."""

    file_location = None

    def __init__(self, app: App):
        """Upload Disk Driver Constructor.

        Arguments:
            StorageConfig {config.storage} -- Storage configuration.
            Application {masonite.app.App} -- The application container.
        """
        self.config = app.make('StorageConfig')
        self.appconfig = app.make('Application')

    def store(self, fileitem, filename=None, location=None):
        """Store the file onto a server.

        Arguments:
            fileitem {cgi.Storage} -- Storage object.

        Keyword Arguments:
            location {string} -- The location on disk you would like to store the file. (default: {None})
            filename {string} -- A new file name you would like to name the file. (default: {None})

        Returns:
            string -- Returns the file name just saved.
        """

        # use the new filename or get it from the fileitem
        if filename is None:
            filename = self.get_name(fileitem)

        # Check if is a valid extension
        self.validate_extension(filename)

        location = self.get_location(location)
        if not location.endswith('/'):
            location += '/'

        if isinstance(fileitem, _io.TextIOWrapper):
            open(location + filename, 'wb').write(bytes(fileitem.read(), 'utf-8'))
        else:
            open(location + filename, 'wb').write(fileitem.file.read())

        self.file_location = location + filename

        return filename



from abc import ABC, abstractmethod


class SessionContract(ABC):

    @abstractmethod
    def get(self):
        pass

    @abstractmethod
    def set(self):
        pass

    @abstractmethod
    def has(self):
        pass

    @abstractmethod
    def all(self):
        pass

    @abstractmethod
    def delete(self):
        pass

    @abstractmethod
    def flash(self):
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def helper(self):
        pass



import io

from masonite.app import App
from masonite.request import Request
from masonite.testsuite import TestSuite, generate_wsgi


class MockRoute:

    def __init__(self, route, container):
        self.route = route
        self.container = container

    def is_named(self, name):
        return self.route.named_route == name

    def has_middleware(self, *middleware):
        return all(elem in self.route.list_middleware for elem in middleware)

    def has_controller(self, controller):
        return self.route.controller == controller

    def contains(self, value):
        wsgi = generate_wsgi()
        wsgi['PATH_INFO'] = self.route.route_url
        wsgi['REQUEST_METHOD'] = self.route.method_type[0]
        self.container = self._run_container(wsgi).container

        return value in self.container.make('Response')

    def status(self, value=None):
        wsgi = generate_wsgi()
        wsgi['PATH_INFO'] = self.route.route_url
        wsgi['REQUEST_METHOD'] = self.route.method_type[0]
        self.container = self._run_container(wsgi).container

        if not value:
            return self.container.make('Request').get_status_code()

        return self.container.make('Request').get_status_code() == value

    def ok(self):
        return self.status('200 OK')

    def can_view(self):
        wsgi = generate_wsgi()
        wsgi['PATH_INFO'] = self.route.route_url
        wsgi['RAW_URI'] = self.route.route_url
        self.container = self._run_container(wsgi).container

        return self.container.make('Request').get_status_code() == '200 OK'

    def user(self, obj):
        self._user = obj
        self.container.on_resolve(Request, self._bind_user_to_request)
        return self

    def is_post(self):
        return 'POST' in self.route.method_type

    def is_get(self):
        return 'GET' in self.route.method_type

    def is_put(self):
        return 'PUT' in self.route.method_type

    def is_patch(self):
        return 'PATCH' in self.route.method_type

    def is_delete(self):
        return 'DELETE' in self.route.method_type

    def on_bind(self, obj, method):
        self.container.on_bind(obj, method)
        return self

    def has_session(self, key):
        wsgi = generate_wsgi()
        wsgi['PATH_INFO'] = self.route.route_url
        wsgi['RAW_URI'] = self.route.route_url
        self.container = self._run_container(wsgi).container
        return self.container.make('Session').has(key)

    def session(self, key):
        wsgi = generate_wsgi()
        wsgi['PATH_INFO'] = self.route.route_url
        wsgi['RAW_URI'] = self.route.route_url
        self.container = self._run_container(wsgi).container
        return self.container.make('Session').get(key)

    def on_make(self, obj, method):
        self.container.on_make(obj, method)
        return self

    def on_resolve(self, obj, method):
        self.container.on_resolve(obj, method)
        return self

    def _run_container(self, wsgi):
        return TestSuite().create_container(wsgi, container=self.container)

    def _bind_user_to_request(self, request, container):
        request.set_user(self._user)
        return self



""" Driver for AMQP support """

import pickle
import threading

from config import queue
from masonite.contracts import QueueContract
from masonite.drivers import BaseDriver
from masonite.app import App
from masonite.exceptions import DriverLibraryNotFound

if 'amqp' in queue.DRIVERS:
    listening_channel = queue.DRIVERS['amqp']['channel']
else:
    listening_channel = 'default'


class QueueAmqpDriver(QueueContract, BaseDriver):

    def __init__(self):
        """Queue AMQP Driver

        Arguments:
            Container {masonite.app.App} -- The application container.
        """

        # Start the connection
        self._connect()

    def _connect(self):
        try:
            import pika
            self.pika = pika
        except ImportError:
            raise DriverLibraryNotFound(
                "Could not find the 'pika' library. Run pip install pika to fix this.")

        connection = pika.BlockingConnection(pika.URLParameters('amqp://{}:{}@{}{}/{}'.format(
            queue.DRIVERS['amqp']['username'],
            queue.DRIVERS['amqp']['password'],
            queue.DRIVERS['amqp']['host'],
            ':'
            + str(queue.DRIVERS['amqp']['port']) if 'port' in queue.DRIVERS['amqp'] and queue.DRIVERS['amqp']['port'] else '',
            queue.DRIVERS['amqp']['vhost'] if 'vhost' in queue.DRIVERS['amqp'] and queue.DRIVERS['amqp']['vhost'] else '%2F'
        )))

        # Get the channel
        self.channel = connection.channel()

        # Declare what queue we are working with
        self.channel.queue_declare(queue=listening_channel, durable=True)

    def _publish(self, body):
        self.channel.basic_publish(exchange='',
                                   routing_key=listening_channel,
                                   body=pickle.dumps(
                                       body
                                   ),
                                   properties=self.pika.BasicProperties(
                                       delivery_mode=2,  # make message persistent
                                   ))

    def push(self, *objects, args=()):
        """Push objects onto the amqp stack.

        Arguments:
            objects {*args of objects} - This can be several objects as parameters into this method.
        """

        for obj in objects:
            # Publish to the channel for each object
            try:
                self._publish({'obj': obj, 'args': args})
            except self.pika.exceptions.ConnectionClosed:
                self._connect()
                self._publish({'obj': obj, 'args': args})



"""A Mail Service Provider."""

from config import mail
from masonite.drivers import MailMailgunDriver, MailSmtpDriver, \
    MailLogDriver, MailTerminalDriver
from masonite.managers import MailManager
from masonite.provider import ServiceProvider
from masonite import Mail


class MailProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('MailConfig', mail)
        self.app.bind('MailSmtpDriver', MailSmtpDriver)
        self.app.bind('MailMailgunDriver', MailMailgunDriver)
        self.app.bind('MailLogDriver', MailLogDriver)
        self.app.bind('MailTerminalDriver', MailTerminalDriver)
        self.app.bind('MailManager', MailManager(self.app))

    def boot(self, manager: MailManager):
        self.app.bind('Mail', manager.driver(self.app.make('MailConfig').DRIVER))
        self.app.swap(Mail, manager.driver(self.app.make('MailConfig').DRIVER))



"""Migrate Reset Command."""
import os
import sys

from cleo import Command
from masonite.packages import add_venv_site_packages
from orator.exceptions.query import QueryException


class MigrateResetCommand(Command):
    """
    Migrate reset.

    migrate:reset
    """

    def handle(self):
        sys.path.append(os.getcwd())
        try:
            add_venv_site_packages()
            from wsgi import container
        except ImportError:
            self.comment(
                'This command must be ran inside of the root of a Masonite project directory')

        # Get any migration files from the Service Container
        migration_directory = ['databases/migrations']
        for key, value in container.providers.items():
            if type(key) == str and 'MigrationDirectory' in key:
                migration_directory.append(value)

        # Load in the Orator migration system
        from orator.migrations import Migrator, DatabaseMigrationRepository
        from config import database
        repository = DatabaseMigrationRepository(database.DB, 'migrations')
        migrator = Migrator(repository, database.DB)
        if not migrator.repository_exists():
            repository.create_repository()

        # Create a new list of migrations with the correct file path instead
        migration_list = []
        for migration in migrator.get_repository().get_ran():
            for directory in migration_directory:
                if os.path.exists(os.path.join(directory, migration + '.py')):
                    migration_list.append(os.path.join(os.getcwd(), directory))
                    break

        # Rollback the migrations
        notes = []
        for migration in migrator.get_repository().get_ran():
            for migration_directory in migration_list:
                try:
                    migrator.reset(migration_directory)
                except QueryException as e:
                    raise e
                except FileNotFoundError:
                    pass

                if migrator.get_notes():
                    notes += migrator.get_notes()

        # Show notes from the migrator
        self.line('')
        for note in notes:
            if not ('Nothing to rollback.' in note):
                self.line(note)
        if not notes:
            self.info('Nothing to rollback')



"""Displays Information Command."""
import math
import os
import platform
import sys
import psutil

from cleo import Command
from tabulate import tabulate

from masonite.info import VERSION


class InfoCommand(Command):
    """
    Displays environment info for debugging.

    info
    """

    def handle(self):
        from masonite_cli.application import application
        rows = []

        rows.append(['System Information', self._get_system_info()])
        mem = math.ceil(psutil.virtual_memory().total / 1024 / 1024 / 1024.0)
        rows.append(['System Memory', str(mem) + ' GB'])
        rows.append(['Python Version', self._get_python_info()])
        rows.append(['Virtual Environment', self._check_virtual_environment()])
        rows.append(['Masonite Version', VERSION])
        rows.append(['Craft Version', application._version])

        if 'APP_ENV' in os.environ:
            rows.append(['APP_ENV', os.environ.get('APP_ENV')])

        if 'APP_DEBUG' in os.environ:
            rows.append(['APP_DEBUG', os.environ.get('APP_DEBUG')])

        self.info('')
        self.info(tabulate(rows, headers=['Environment Information', '']))
        self.info('')

    def _get_python_info(self):
        py_version = platform.python_version()
        py_implementation = platform.python_implementation()
        return py_implementation + ' ' + py_version

    def _check_virtual_environment(self):
        if hasattr(sys, 'real_prefix') or 'VIRTUAL_ENV' in os.environ:
            return u'\u2713'  # currently running in virtual env
        return 'X'

    def _get_system_info(self):
        bits, _ = platform.architecture()
        operating_system, _, _, _, arch, _ = platform.uname()

        if operating_system.lower() == 'darwin':
            operating_system = 'MacOS'
        return '{} {} {}'.format(operating_system, arch, bits)



from .managers.BroadcastManager import Broadcast
from .managers.CacheManager import Cache
from .managers.MailManager import Mail
from .managers.QueueManager import Queue
from .managers.SessionManager import Session
from .managers.UploadManager import Upload
from masonite.environment import env



"""An Upload Service Provider."""

from config import storage
from masonite.drivers import UploadDiskDriver, UploadS3Driver
from masonite.helpers.static import static
from masonite.managers import UploadManager
from masonite.provider import ServiceProvider
from masonite.view import View
from masonite import Upload


class UploadProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('StorageConfig', storage)
        self.app.bind('UploadDiskDriver', UploadDiskDriver)
        self.app.bind('UploadS3Driver', UploadS3Driver)
        self.app.bind('UploadManager', UploadManager(self.app))

    def boot(self, manager: UploadManager, view: View):
        self.app.bind('Upload', manager.driver(self.app.make('StorageConfig').DRIVER))
        self.app.swap(Upload, manager.driver(self.app.make('StorageConfig').DRIVER))
        view.share(
            {
                'static': static,
            }
        )



"""A SassProvider Service Provider."""

import builtins
import os

from masonite.exception_handler import DD, DumpHandler
from masonite.helpers.view_helpers import back, set_request_method
from masonite.provider import ServiceProvider
from masonite.view import View
from masonite.request import Request
from masonite.managers import MailManager


class HelpersProvider(ServiceProvider):

    wsgi = False

    def register(self):
        pass

    def boot(self, view: View, request: Request):
        """Add helper functions to Masonite."""
        builtins.view = view.render
        builtins.request = request.helper
        builtins.auth = request.user
        builtins.container = self.app.helper
        builtins.env = os.getenv
        builtins.resolve = self.app.resolve
        builtins.route = request.route
        if self.app.has(MailManager):
            builtins.mail_helper = self.app.make(MailManager).helper
        builtins.dd = DD(self.app).dump

        view.share(
            {
                'request': request.helper,
                'auth': request.user,
                'request_method': set_request_method,
                'route': request.route,
                'back': back
            }
        )



from .BaseRequest import BaseRequest
from .MockJson import MockJson
from .MockRequest import MockRequest
from .MockRoute import MockRoute
from .UnitTest import UnitTest
from .MockController import MockController



"""Cache Manager."""

from masonite.contracts import CacheManagerContract
from masonite.managers import Manager


class CacheManager(Manager, CacheManagerContract):
    """Manages all cache drivers.

    Arguments:
        Manager {masonite.managers.Manager} -- The base Manager class.
    """

    config = 'CacheConfig'
    driver_prefix = 'Cache'


class Cache:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



"""View Helper Module."""

from jinja2 import Markup


def set_request_method(method_type):
    """Return an input string for use in a view to change the request method of a form.

    Arguments:
        method_type {string} -- Can be options like GET, POST, PUT, PATCH, DELETE

    Returns:
        string -- An input string.
    """
    return Markup("<input type='hidden' name='__method' value='{0}'>".format(method_type))


def back(location):
    """Return an input element for use in telling Masonite which route to redirect back to.

    Arguments:
        location {string} -- The route to redirect back to.

    Returns:
        string -- An input string.
    """
    return Markup("<input type='hidden' name='__back' value='{0}'>".format(location))



"""Upload Manager Module."""

from masonite.contracts import UploadManagerContract
from masonite.managers import Manager


class UploadManager(Manager, UploadManagerContract):
    """Manages all upload drivers.

    Arguments:
        Manager {masonite.managers.Manager} -- The base Manager class.
    """

    config = 'StorageConfig'
    driver_prefix = 'Upload'


class Upload:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



"""New Seeder Command."""
import subprocess
import os

from cleo import Command


class SeedCommand(Command):
    """
    Create a seeder to seed a database.

    seed
        {table : Name of the table to seed}
    """

    def handle(self):
        table = self.argument('table').lower()
        subprocess.call([
            "orator make:seed {}_table_seeder -p databases/seeds".format(
                table),
        ], shell=True)

        self.check_init_file()

    def check_init_file(self):
        os.makedirs(os.path.dirname(
            'databases/seeds/__init__.py'), exist_ok=True)

        if 'sys.path.append(os.getcwd())' not in open('databases/seeds/__init__.py').read():
            f = open('databases/seeds/__init__.py', 'w+')
            f.write('import os\nimport sys\nsys.path.append(os.getcwd())\n')
            f.close()



from .Manager import Manager
from .BroadcastManager import BroadcastManager
from .CacheManager import CacheManager
from .MailManager import MailManager
from .QueueManager import QueueManager
from .SessionManager import SessionManager
from .UploadManager import UploadManager



"""The ConfirmController Module."""
import time
import datetime

from masonite.auth import Auth
from masonite.auth.Sign import Sign
from masonite.request import Request
from masonite.view import View
from masonite.auth import MustVerifyEmail
from app.User import User


class ConfirmController:
    """The ConfirmController class."""

    def __init__(self):
        """The ConfirmController Constructor."""
        pass

    def verify_show(self, request: Request, view: View):
        """Show the Verify Email page for unverified users.

        Arguments:
            Request {masonite.request.request} -- The Masonite request class.
            Request {masonite.view.view} -- The Masonite view class.

        Returns:
            [type] -- [description]
        """
        return view.render('auth/verify', {'app': request.app().make('Application'), 'Auth': Auth(request)})

    def confirm_email(self, request: Request, view: View):
        """Confirm User email and show the correct response.

        Arguments:
            Request {masonite.request.request} -- The Masonite request class.
            Request {masonite.view.view} -- The Masonite view class.

        Returns:
            [type] -- [description]
        """
        sign = Sign()
        token = sign.unsign(request.param('id'))

        if token is not None:
            tokenParts = token.split("::")
            if len(tokenParts) > 1:
                id = tokenParts[0]
                user = self.get_user(id)

                if user.verified_at is None:
                    timestamp = datetime.datetime.fromtimestamp(float(tokenParts[1]))
                    now = datetime.datetime.now()
                    timestamp_plus_10 = timestamp + datetime.timedelta(minutes=10)

                    if now < timestamp_plus_10:
                        user.verified_at = datetime.datetime.now()
                        user.save()

                        return view.render('auth/confirm', {'app': request.app().make('Application'), 'Auth': Auth(request)})

        return view.render('auth/error', {'app': request.app().make('Application'), 'Auth': Auth(request)})

    def get_user(self, id):
        """Get the user from the database.

        Arguments:
            id {str} -- The user id

        Returns:
            [User] -- [User model]
        """
        return User.find(id)

    def send_verify_email(self, request: Request):
        user = request.user()

        if isinstance(user, MustVerifyEmail):
            request.app().resolve(user.verify_email)

        return request.redirect('/home')



"""New Migration Command."""
import subprocess

from cleo import Command


class MakeMigrationCommand(Command):
    """
    Makes a new migration.

    migration
        {name : Name of your migration}
        {--t|--table=False : Table you are migrating for}
        {--c|--create=False : Table you want to create with this migration}
    """

    def handle(self):
        name = self.argument('name')

        if self.option('create') != 'False':
            subprocess.call(['orator', 'make:migration', name,
                             '-p', 'databases/migrations', '--table', self.option('create'), '--create'])
        elif self.option('table') != 'False':
            subprocess.call(['orator', 'make:migration', name,
                             '-p', 'databases/migrations', '--table', self.option('table')])
        else:
            subprocess.call(['orator', 'make:migration', name,
                             '-p', 'databases/migrations'])



"""Validation Module."""

import validator


class Validator:
    """Validator Class. Responsible for validating form data and dictionaries."""

    def __init__(self, request=None):
        """Validator constructor.

        Keyword Arguments:
            request {masonite.request.Request} -- Request class. (default: {None})
        """
        self.request = request
        self.validation_dictionary = {}
        self.check_manual_dictionary = False
        self.error_messages = {}

    def validate(self, dictionary):
        """Set the validation dictionary.

        Arguments:
            dictionary {dict} -- Dictionary to validate

        Returns:
            self
        """
        self.validation_dictionary = dictionary
        return self

    def check(self, check_manual_dictionary=False):
        """Validate the dictionary.

        Keyword Arguments:
            check_manual_dictionary {bool} -- Check a dictionary manually. (default: {False})

        Returns:
            bool|dict
        """
        if check_manual_dictionary:
            self.check_manual_dictionary = check_manual_dictionary
        return self.run_validation()[0]

    def errors(self):
        """Return a dictionary of errors.

        Returns:
            None|dict
        """
        validation = self.run_validation()
        for message in self.error_messages:
            if message in validation[1]:
                validation[1][message] = self.error_messages[message]

        if not validation[1]:
            return None

        return validation[1]

    def run_validation(self):
        """Load the dictionary and runs the validations.

        Returns:
            bool|dict
        """
        validation_dict = self.load_request_input()
        for validation in validation_dict:
            validation_dict.update({
                validation: self.get(validation)
            })

        return validator.validate(self.validation_dictionary, validation_dict)

    def get(self, validation):
        """Get a key in the validation input and runs it through a cast method if one exists.

        Arguments:
            validation {string} -- Key inside the validation input

        Returns:
            string -- Returns the validation
        """
        if hasattr(self, "cast_{}".format(validation)):
            validation_input = self.load_request_input()[validation]
            return getattr(self, "cast_{}".format(validation))(validation_input)

        return self.load_request_input()[validation]

    def error(self, error_id):
        """Get an error from the error dictionary by key.

        Arguments:
            error_id {string} -- Error key to search for.

        Returns:
            string|None
        """
        if error_id in self.errors():
            return self.errors()[error_id]

        return None

    def load_request_input(self):
        """Find which input to load into the validation.

        Returns:
            dict -- Returns the dictionary of values to validate.
        """
        if self.request:
            return self.request.all().copy()

        return self.check_manual_dictionary

    def messages(self, messages):
        """Specify custom error messages if the validation fails.

        Arguments:
            messages {dict} -- Sets a dictionary of error messages.
        """
        self.error_messages = messages



"""A Csrf Service Provider."""

from masonite.auth import Csrf
from masonite.provider import ServiceProvider


class CsrfProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('Csrf', Csrf(self.app.make('Request')))

    def boot(self):
        pass



from cleo import Command

from masonite.app import App
from masonite.helpers.filesystem import make_directory
from masonite.view import View


class BaseScaffoldCommand(Command):
    """
    Creates a model.

    model
        {name : Name of the model}
    """

    scaffold_name = 'Example'
    suffix = ''
    file_extension = '.py'
    base_directory = 'app/example/'
    template = '/masonite/snippets/scaffold/model'

    def handle(self):
        class_name = self.argument('name')
        view = View(App())
        class_directory = '{}{}{}{}'.format(
            self.base_directory, class_name, self.suffix, self.file_extension)

        if not make_directory(class_directory):
            return self.error('{0} Already Exists!'.format(self.scaffold_name))

        f = open(class_directory, 'w+')
        if view.exists(self.template):
            f.write(
                view.render(self.template, {
                            'class': class_name.split('/')[-1]}).rendered_template
            )
            self.info('{} Created Successfully!'.format(self.scaffold_name))
            return f.close()



from masonite.routes import Get
from routes import web


class TestRoute:

    def __init__(self, route):
        self.route = None

        for routes in web.ROUTES:
            if routes.route_url == route:
                self.route = routes
                break

    def exists(self):
        return isinstance(self.route, Get)

    def has_middleware(self, middleware):
        return middleware in self.route.list_middleware



"""Base Queue Module."""


class Queueable:
    """Makes classes Queueable."""

    def handle(self):
        """Put the queue logic in this handle method."""
        pass

    def dispatch(self):
        """Responsible for dispatching the job to the Queue service.

        Returns:
            self.handle
        """
        return self.handle



import os


def make_directory(directory):
    if not os.path.isfile(directory):
        if not os.path.exists(os.path.dirname(directory)):
            # Create the path to the model if it does not exist
            os.makedirs(os.path.dirname(directory))

        return True

    return False



"""New Middleware Command."""
from masonite.commands import BaseScaffoldCommand


class MiddlewareCommand(BaseScaffoldCommand):
    """
    Creates a middleware.

    middleware
        {name : Name of the middleware}
    """

    scaffold_name = "Middleware"
    suffix = "Middleware"
    template = "/masonite/snippets/scaffold/middleware"
    base_directory = "app/http/middleware/"



"""The HomeController Module."""

from masonite.auth import Auth
from masonite.request import Request
from masonite.view import View


class HomeController:
    """Home Dashboard Controller."""

    def __init__(self):
        pass

    def show(self, request: Request, view: View):
        if not Auth(request).user():
            request.redirect('/login')
        return view.render('auth/home', {'app': request.app().make('Application'), 'Auth': Auth(request)})



"""Core of the IOC Container."""

import inspect

from masonite.exceptions import (ContainerError,
                                 MissingContainerBindingNotFound,
                                 StrictContainerException)


class App:
    """Core of the Service Container.

    Performs bindings and resolving of objects to and from the container.
    """

    def __init__(self, strict=False, override=True, resolve_parameters=False):
        """App class constructor."""
        self.providers = {}
        self.strict = strict
        self.override = override
        self.resolve_parameters = resolve_parameters
        self._hooks = {
            'make': {},
            'bind': {},
            'resolve': {},
        }

        self.swaps = {}

    def bind(self, name, class_obj):
        """Bind classes into the container with a key value pair.

        Arguments:
            name {string} -- Name of the key you want to bind the object to
            class_obj {object} -- The object you want to bind

        Returns:
            self
        """
        if self.strict and name in self.providers:
            raise StrictContainerException(
                'You cannot override a key inside a strict container')

        if self.override or name not in self.providers:
            self.fire_hook('bind', name, class_obj)
            self.providers.update({name: class_obj})

        return self

    def simple(self, obj):
        """Easy way to bind classes into the container by using passing the object only.

        Automatically generates the key for the binding process.

        Arguments:
            class_obj {object} -- The object you want to bind

        Returns:
            self
        """
        self.bind(obj if inspect.isclass(obj) else obj.__class__, obj)
        return self

    def make(self, name):
        """Retrieve a class from the container by key.

        Arguments:
            name {string} -- Key in the container that you want to get.

        Raises:
            MissingContainerBindingNotFound -- Raised if the key is not in the container.

        Returns:
            object -- Returns the object that is fetched.
        """
        if name in self.providers:
            obj = self.providers[name]
            self.fire_hook('make', name, obj)
            return obj
        elif name in self.swaps:
            return self.swaps[name]
        elif inspect.isclass(name):
            obj = self._find_obj(name)
            self.fire_hook('make', name, obj)
            return obj

        raise MissingContainerBindingNotFound(
            "{0} key was not found in the container".format(name))

    def has(self, name):
        """Check if a key exists in the container.

        Arguments:
            name {string} -- Key you want to check for in the container

        Returns:
            bool
        """
        if isinstance(name, str):
            return name in self.providers
        else:
            try:
                self._find_obj(name)
                return True
            except MissingContainerBindingNotFound:
                return False

        return False

    def helper(self):
        """Add a helper to create builtin functions.

        Used to more simply return
        instances of this class when building helpers.

        Returns:
            self
        """
        return self

    def resolve(self, obj, *resolving_arguments):
        """Takes an object such as a function or class method and resolves it's
        parameters from objects in the container.

        Arguments:
            obj {object} -- The object you want to resolve objects from via this container.

        Returns:
            object -- The object you tried resolving but with the correct dependencies injected.
        """
        provider_list = []
        passing_arguments = list(resolving_arguments)

        for _, value in inspect.signature(obj).parameters.items():
            if ':' in str(value):
                provider_list.append(self._find_annotated_parameter(value))
            elif self.resolve_parameters:
                provider_list.append(self._find_parameter(value))
            elif resolving_arguments:
                try:
                    provider_list.append(passing_arguments.pop(0))
                except IndexError:
                    raise ContainerError('Not enough dependencies passed. Resolving object needs {} dependencies.'.format(len(inspect.signature(obj).parameters)))
            else:
                raise ContainerError(
                    "This container is not set to resolve parameters. You can set this in the container"
                    " constructor using the 'resolve_parameters=True' keyword argument.")
        try:
            return obj(*provider_list)
        except TypeError as e:
            raise ContainerError('Either Tried resolving the too many dependencies or {}'.format(str(e)))

    def collect(self, search):
        """Fetch a dictionary of objects using a search query.

        Arguments:
            search {string|object} -- The string or object you want to search for.

        Raises:
            AttributeError -- Thrown if there is no wildcard in the search string

        Returns:
            dict -- Returns a dictionary of collected objects and their key bindings.
        """
        provider_list = {}
        if isinstance(search, str):
            # Search Can Be:
            #    '*ExceptionHook'
            #    'Sentry*'
            #    'Sentry*Hook'
            for key, value in self.providers.items():
                if isinstance(key, str):
                    if search.startswith('*'):
                        if key.endswith(search.split('*')[1]):
                            provider_list.update({key: value})
                    elif search.endswith('*'):
                        if key.startswith(search.split('*')[0]):
                            provider_list.update({key: value})
                    elif '*' in search:
                        split_search = search.split('*')
                        if key.startswith(split_search[0]) and key.endswith(split_search[1]):
                            provider_list.update({key: value})
                    else:
                        raise AttributeError(
                            "There is no '*' in your collection search")
        else:
            for provider_key, provider_class in self.providers.items():
                if inspect.isclass(provider_class) and issubclass(provider_class, search):
                    provider_list.update({provider_key: provider_class})

        return provider_list

    def _find_annotated_parameter(self, parameter):
        """Find a given annotation in the container.

        Arguments:
            parameter {object} -- The object to find in the container.

        Raises:
            ContainerError -- Thrown when the dependency is not found in the container.

        Returns:
            object -- Returns the object found in the container.
        """
        if parameter.annotation in self.swaps:
            obj = self.swaps[parameter.annotation]
            if callable(obj):
                return self.swaps[parameter.annotation](parameter.annotation, self)
            return obj

        for _, provider_class in self.providers.items():

            if parameter.annotation == provider_class or parameter.annotation == provider_class.__class__:
                obj = provider_class
                self.fire_hook('resolve', parameter, obj)

                return obj
            elif inspect.isclass(provider_class) and issubclass(provider_class, parameter.annotation) or issubclass(provider_class.__class__, parameter.annotation):
                obj = provider_class
                self.fire_hook('resolve', parameter, obj)
                return obj

        raise ContainerError(
            'The dependency with the {0} annotation could not be resolved by the container'.format(parameter))

    def _find_parameter(self, parameter):
        """Find a parameter in the container.

        Arguments:
            parameter {string} -- Parameter to search for.

        Raises:
            ContainerError -- Thrown when the dependency is not found in the container.

        Returns:
            object -- Returns the object found in the container
        """
        parameter = str(parameter)
        if parameter is not 'self' and parameter in self.providers:
            obj = self.providers[parameter]
            self.fire_hook('resolve', parameter, obj)
            return obj
        raise ContainerError(
            'The dependency with the key of {0} could not be found in the container'.format(
                parameter)
        )

    def on_bind(self, key, obj):
        """Set some listeners for when a specific key or class in binded to the container.

        Arguments:
            key {string|object} -- The key can be a string or an object to listen for
            obj {object} -- Should be a function or class method

        Returns:
            self
        """
        return self._bind_hook('bind', key, obj)

    def on_make(self, key, obj):
        """Set some listeners for when a specific key or class is made from the container.

        Arguments:
            key {string|object} -- The key can be a string or an object to listen for
            obj {object} -- Should be a function or class method

        Returns:
            self
        """
        return self._bind_hook('make', key, obj)

    def on_resolve(self, key, obj):
        """Set some listeners for when a specific key or class is resolved from the container.

        Arguments:
            key {string|object} -- The key can be a string or an object to listen for
            obj {object} -- Should be a function or class method

        Returns:
            self
        """
        return self._bind_hook('resolve', key, obj)

    def swap(self, obj, callback):
        self.swaps.update({obj: callback})
        return self

    def fire_hook(self, action, key, obj):
        """Fire a specific hook based on a key or object.

        Arguments:
            action {string} -- Should be the action to fire (bind|make|resolve)
            key {string|object} -- The key can be a string or an object to listen for
            obj {object} -- Should be a function or class method

        Returns:
            None
        """
        if str(key) in self._hooks[action] or \
                inspect.isclass(obj) and \
                obj in self._hooks[action] or obj.__class__ in self._hooks[action]:

            for hook, hook_list in self._hooks[action].items():
                for hook_obj in hook_list:
                    hook_obj(obj, self)

    def _bind_hook(self, hook, key, obj):
        """Internal method used to abstract away the logic for binding an listener to the container hooks.

        Arguments:
            hook {string} -- The hook you want to listen for (bind|make|resolve)
            key {string|object} -- The key to save for the listener
            obj {object} -- Should be a function or class method

        Returns:
            self
        """
        if key in self._hooks[hook]:
            self._hooks[hook][key].append(obj)
        else:
            self._hooks[hook].update({key: [obj]})
        return self

    def _find_obj(self, obj):
        """Find an object in the container.

        Arguments:
            obj {object} -- Any object in the container

        Raises:
            MissingContainerBindingNotFound -- Raised when the object cannot be found.

        Returns:
            object -- Returns the object in the container
        """
        for _, provider_class in self.providers.items():
            if obj == provider_class or obj == provider_class.__class__:
                return_obj = provider_class
                self.fire_hook('resolve', obj, return_obj)
                return return_obj
            elif inspect.isclass(provider_class) and issubclass(provider_class, obj) or issubclass(provider_class.__class__, obj):
                return_obj = provider_class
                self.fire_hook('resolve', obj, return_obj)
                return return_obj

        raise MissingContainerBindingNotFound(
            'The dependency with the {0} annotation could not be resolved by the container'.format(obj))



"""The Masonite Response Object."""

import json

from masonite.exceptions import ResponseError
from masonite.helpers.Extendable import Extendable
from masonite.view import View


class Response(Extendable):

    def __init__(self, app):
        """A Response object to be used to abstract the logic of getting a response ready to be returned.

        Arguments:
            app {masonite.app.App} -- The Masonite container.
        """
        self.app = app
        self.request = self.app.make('Request')

    def json(self, payload):
        """Gets the response ready for a JSON response.

        Arguments:
            payload {dict|list} -- Either a dictionary or a list.

        Returns:
            string -- Returns a string representation of the data
        """
        self.app.bind('Response', json.dumps(payload))
        self.make_headers(content_type="application/json; charset=utf-8")
        self.request.status(200)

        return self.data()

    def make_headers(self, content_type="text/html; charset=utf-8"):
        """Make the appropriate headers based on changes made in controllers or middleware.

        Keyword Arguments:
            content_type {str} -- The content type to set. (default: {"text/html; charset=utf-8"})
        """
        self.request.header('Content-Length', str(len(self.data())))

        # If the user did not change it directly
        if not self.request.header('Content-Type'):
            self.request.header('Content-Type', content_type)

    def data(self):
        """Get the data that will be returned to the WSGI server.

        Returns:
            string -- Returns a string representation of the response
        """
        if self.app.has('Response'):
            return self.app.make('Response')

        return ''

    def converted_data(self):
        """Converts the data appropriately so the WSGI server can handle it.

        Returns:
            string -- Returns a string representation of the data
        """
        if isinstance(self.data(), dict) or isinstance(self.data(), list):
            return json.dumps(self.data())
        else:
            return self.data()

    def view(self, view, status=200):
        """Set a string or view to be returned.

        Arguments:
            view {string|dict|list|masonite.view.View} -- Some data type that is an appropriate response.

        Keyword Arguments:
            status {int} -- The Response status code. (default: {200})

        Raises:
            ResponseError -- If a data type that is not an acceptable response type is returned.

        Returns:
            string|dict|list -- Returns the data to be returned.
        """
        self.request.status(status)

        if isinstance(view, dict) or isinstance(view, list):
            return self.json(view)
        elif isinstance(view, View):
            view = view.rendered_template
        elif isinstance(view, self.request.__class__):
            view = self.data()
        elif view is None:
            raise ResponseError('Responses cannot be of type: None.')

        self.app.bind('Response', view)

        self.make_headers()

        return self.data()

    def redirect(self, location=None, status=302):
        """Set the redirection on the server.

        Keyword Arguments:
            location {string} -- The URL to redirect to (default: {None})
            status {int} -- The Response status code. (default: {302})

        Returns:
            string -- Returns the data to be returned.
        """
        self.request.status(status)
        if not location:
            location = self.request.redirect_url

        self.request.header('Location', location)
        self.app.bind('Response', 'redirecting ...')

        return self.data()

    def to_bytes(self):
        """Converts the data to bytes so the WSGI server can handle it.

        Returns:
            bytes -- The converted response to bytes.
        """
        return bytes(self.converted_data(), 'utf-8')



"""New View Command."""
from masonite.commands import BaseScaffoldCommand


class ViewCommand(BaseScaffoldCommand):
    """
    Creates a view.

    view
        {name : Name of the view you would like to create}
    """

    scaffold_name = "View"
    template = '/masonite/snippets/scaffold/view'
    file_extension = '.html'
    base_directory = 'resources/templates/'



"""Install Command."""
import os
import shutil
import subprocess

from cleo import Command


class InstallCommand(Command):
    """
    Installs all of Masonite's dependencies.

    install
    """

    def handle(self):
        subprocess.call(["pip3", "install", "-r", "requirements.txt"])

        # create the .env file if it does not exist
        if not os.path.isfile('.env'):
            shutil.copy('.env-example', '.env')

        subprocess.call(["craft", "key", "--store"])



"""Base upload driver module."""

from masonite.exceptions import FileTypeException
from masonite.drivers.BaseDriver import BaseDriver
from masonite.helpers import random_string
import _io


class BaseUploadDriver(BaseDriver):
    """Base class that all upload drivers inherit from."""

    accept_file_types = None

    def accept(self, *args, **kwargs):
        """Set file types to accept before uploading.

        Returns:
            self
        """
        self.accept_file_types = args
        return self

    def validate_extension(self, filename):
        """Check for valid file extenstions set with the 'accept' method.

        Arguments:
            filename {string} -- The filename with file extension to validate.

        Raises:
            FileTypeException -- Thrown if the specified file extension is incorrect.
        """
        if self.accept_file_types is not None:
            if not filename.endswith(self.accept_file_types):
                raise FileTypeException("The extension file not is valid.")

    def get_location(self, location=None):
        """Get the location of where to upload.

        Keyword Arguments:
            location {string} -- The path to upload to. If none then this will check for configuration settings. (default: {None})

        Returns:
            string -- Returns the location it uploaded to.
        """
        if not location:
            location = self.config.DRIVERS['disk']['location']

        if '.' in location:
            location = location.split('.')
            return self.config.DRIVERS[location[0]]['location'][location[1]]
        elif isinstance(location, str):
            return location
        elif isinstance(location, dict):
            return list(location.values())[0]

        return location

    def get_name(self, fileitem):
        return "{}.{}".format(random_string(25).lower(), self.get_extension(fileitem))

    def get_extension(self, fileitem):
        if isinstance(fileitem, _io.TextIOWrapper):
            return fileitem.name.split('.')[-1]
        else:
            return fileitem.filename.split('.')[-1]



"""Helper Functions for RouteProvider."""

import re


def flatten_routes(routes):
    """Flatten the grouped routes into a single list of routes.

    Arguments:
        routes {list} -- This can be a multi dementional list which can flatten all lists into a single list.

    Returns:
        list -- Returns the flatten list.
    """
    route_collection = []
    for route in routes:
        # Check if a route is a list of routes
        if isinstance(route, list):
            for r in flatten_routes(route):
                route_collection.append(r)
        else:
            route_collection.append(route)

    return route_collection


def get(url, controller):
    """Shortcut for Get HTTP class.

    Arguments:
        url {string} -- The url you want to use for the route
        controller {string|object} -- This can be a string controller or a normal object controller

    Returns:
        masonite.routes.Get -- The Masonite Get class.
    """
    from masonite.routes import Get

    return Get().route(url, controller)


def match(method_type, url, controller):
    """Shortcut for Match HTTP class.

    Arguments:
        method_type {list} -- A list of the method types you want to accept
        url {string} -- The url you want to use for the route
        controller {string|object} -- This can be a string controller or a normal object controller

    Returns:
        masonite.routes.Get -- The Masonite Get class.
    """
    from masonite.routes import Match

    return Match(method_type).route(url, controller)


def post(url, controller):
    """Shortcut for Post HTTP class.

    Arguments:
        url {string} -- The url you want to use for the route
        controller {string|object} -- This can be a string controller or a normal object controller

    Returns:
        masonite.routes.Post -- The Masonite Post class.
    """
    from masonite.routes import Post

    return Post().route(url, controller)


def delete(url, controller):
    """Shortcut for Delete HTTP class.

    Arguments:
        url {string} -- The url you want to use for the route
        controller {string|object} -- This can be a string controller or a normal object controller

    Returns:
        masonite.routes.Delete -- The Masonite Delete class.
    """
    from masonite.routes import Delete

    return Delete().route(url, controller)


def put(url, controller):
    """Shortcut for Put HTTP class.

    Arguments:
        url {string} -- The url you want to use for the route
        controller {string|object} -- This can be a string controller or a normal object controller

    Returns:
        masonite.routes.Put -- The Masonite Put class.
    """
    from masonite.routes import Put

    return Put().route(url, controller)


def patch(url, controller):
    """Shortcut for Patch HTTP class.

    Arguments:
        url {string} -- The url you want to use for the route
        controller {string|object} -- This can be a string controller or a normal object controller

    Returns:
        masonite.routes.Patch -- The Masonite Patch class.
    """
    from masonite.routes import Patch

    return Patch().route(url, controller)


def group(url, route_list):
    """Shortcut for GET HTTP class.

    Arguments:
        url {string} -- The url you want to use for the route
        route_list {list} -- A list of routes

    Returns:
        list -- Returns the list route
    """
    for route in route_list:
        route.route_url = url + route.route_url

    return route_list


def compile_route_to_regex(route):
    """Compile a route to regex.

    Arguments:
        route {masonite.routes.Route} -- The Masonite route object

    Returns:
        string -- Returns the regex of the route.
    """
    # Split the route
    split_given_route = route.split('/')

    # compile the provided url into regex
    url_list = []
    regex = '^'
    for regex_route in split_given_route:
        if '*' in regex_route or '@' in regex_route:
            if ':int' in regex_route:
                regex += r'(\d+)'
            elif ':string' in regex_route:
                regex += r'([a-zA-Z]+)'
            else:
                # default
                regex += r'([\w.-]+)'
            regex += r'\/'

            # append the variable name passed @(variable):int to a list
            url_list.append(
                regex_route.replace('@', '').replace(
                    ':int', '').replace(':string', '')
            )
        else:
            regex += regex_route + r'\/'

    if regex.endswith('/') and not route.endswith('/'):
        regex = regex[:-2]

    regex += '$'

    return regex


def create_matchurl(router, route):
    """Create a regex string for router.url to be matched against.

    Arguments:
        router {masonite.routes.Route} -- The Masonite route object
        route {masonite.routes.BaseHttpRoute} -- The current route being executed.

    Returns:
        string -- compiled regex string
    """
    # Compiles the given route to regex
    regex = route.compile_route_to_regex(router)

    if route.route_url.endswith('/'):
        if router.url.endswith('/'):
            matchurl = re.compile(regex.replace(r'\/\/$', r'\/$'))
        else:
            matchurl = re.compile(regex.replace(r'\/\/$', r'$'))
    else:
        if router.url.endswith('/'):
            matchurl = re.compile(regex)
        else:
            matchurl = re.compile(regex.replace(r'\/$', r'$'))

    return matchurl



"""Validator Helper Module."""

from ..validator import Validator


def validate(validations, data, messages={}):
    """Helper function for shorthand validations.

    Arguments:
        validations {dict} -- A dictionary of validations from the validator.py library
        data {dict} -- The data to be validated

    Keyword Arguments:
        messages {dict} -- Optional messages to render if there was a bad validation. (default: {{}})

    Returns:
        Bool|dict -- Returns True if validations are good or a dictionary if there are errors.
    """
    validator = Validator().validate(validations)
    validator.error_messages = messages
    if validator.check(data):
        return True
    else:
        return validator.errors()



"""Module for the Hook class."""

from masonite.app import App


class Hook:
    """Hook class is responsible for finding and firing framework hooks."""

    def __init__(self, app: App):
        """Hook constructor.

        Arguments:
            app {masonite.app.App} -- Container object.
        """
        self._app = app

    def fire(self, search):
        """Find all the classes to be fired with the exception hook search string.

        Arguments:
            search {string} -- The search string to collect classes with.
        """
        for key in self._app.collect(search):
            self._app.make(key).load(self._app)



"""Module for the Pusher websocket driver."""

from masonite.contracts import BroadcastContract
from masonite.drivers import BaseDriver
from masonite.exceptions import DriverLibraryNotFound
from masonite.app import App


class BroadcastPusherDriver(BroadcastContract, BaseDriver):
    """Class for the Pusher websocket driver."""

    def __init__(self, app: App):
        """Pusher driver constructor.

        Arguments:
            BroadcastConfig {config.broadcast} -- Broadcast configuration.
        """
        self.config = app.make('BroadcastConfig')
        self.ssl_message = True

    def ssl(self, boolean):
        """Set whether to send data with SSL enabled.

        Arguments:
            boolean {bool} -- Boolean on whether to set SSL.

        Returns:
            self
        """
        self.ssl_message = boolean
        return self

    def channel(self, channels, message, event='base-event'):
        """Specify which channel(s) you want to send information to.

        Arguments:
            channels {string|list} -- Can be a string for the channel or a list of strings for the channels.
            message {string} -- The message you want to send to the channel(s)

        Keyword Arguments:
            event {string} -- The event you want broadcasted along with your data. (default: {'base-event'})

        Raises:
            DriverLibraryNotFound -- Thrown when pusher is not installed.

        Returns:
            string -- Returns the message sent.
        """
        try:
            import pusher
        except ImportError:
            raise DriverLibraryNotFound(
                'Could not find the "pusher" library. Please pip install this library running "pip install pusher"')

        pusher_client = pusher.Pusher(
            app_id=str(self.config.DRIVERS['pusher']['app_id']),
            key=self.config.DRIVERS['pusher']['client'],
            secret=self.config.DRIVERS['pusher']['secret'],
            ssl=self.ssl_message
        )

        if isinstance(message, str):
            message = {'message': message}

        if isinstance(channels, list):
            for channel in channels:
                pusher_client.trigger(channel, event, message)
        else:
            pusher_client.trigger(channels, event, message)

        return message



"""Module for controllers. Currently not being used."""


class Controller:
    """Base Class for Controllers."""

    def __init__(self):
        """Controller Constructor."""
        self.app = None

    def load_app(self, app):
        """Load the container into the controller.

        Arguments:
            app {masonite.app.App} -- The container object.

        Returns:
            self
        """
        self.app = app
        return self



"""A RedirectionProvider Service Provider."""

from config import session
from masonite.drivers import SessionCookieDriver, SessionMemoryDriver
from masonite.managers import SessionManager
from masonite.provider import ServiceProvider
from masonite.view import View
from masonite.request import Request
from masonite import Session


class SessionProvider(ServiceProvider):

    def register(self):
        self.app.bind('SessionConfig', session)
        self.app.bind('SessionMemoryDriver', SessionMemoryDriver)
        self.app.bind('SessionCookieDriver', SessionCookieDriver)
        self.app.bind('SessionManager', SessionManager(self.app))

    def boot(self, request: Request, view: View, session: SessionManager):
        self.app.bind('Session', session.driver(self.app.make('SessionConfig').DRIVER))
        self.app.swap(Session, session.driver(self.app.make('SessionConfig').DRIVER))
        request.session = self.app.make('Session')

        view.share({
            'session': self.app.make('Session').helper
        })



"""Upload S3 Driver."""

import os

from masonite.contracts import UploadContract
from masonite.drivers import BaseUploadDriver
from masonite.exceptions import DriverLibraryNotFound
from masonite.managers import UploadManager
from masonite.app import App


class UploadS3Driver(BaseUploadDriver, UploadContract):
    """Amazon S3 Upload driver."""

    def __init__(self, upload: UploadManager, app: App):
        """Upload Disk Driver Constructor.

        Arguments:
            UploadManager {masonite.managers.UploadManager} -- The Upload Manager object.
            StorageConfig {config.storage} -- Storage configuration.
        """
        self.upload = upload
        self.config = app.make('StorageConfig')

    def store(self, fileitem, filename=None, location=None):
        """Store the file into Amazon S3 server.

        Arguments:
            fileitem {cgi.Storage} -- Storage object.

        Keyword Arguments:
            location {string} -- The location on disk you would like to store the file. (default: {None})
            filename {string} -- A new file name you would like to name the file. (default: {None})

        Raises:
            DriverLibraryNotFound -- Raises when the boto3 library is not installed.

        Returns:
            string -- Returns the file name just saved.
        """

        driver = self.upload.driver('disk')
        driver.store(fileitem, location)
        file_location = driver.file_location

        # use the new filename or get it from the fileitem
        if filename is None:
            filename = self.get_name(fileitem)

        # Check if is a valid extension
        self.validate_extension(filename)

        try:
            import boto3
        except ImportError:
            raise DriverLibraryNotFound(
                'Could not find the "boto3" library. Please pip install this library by running "pip install boto3"')

        session = boto3.Session(
            aws_access_key_id=self.config.DRIVERS['s3']['client'],
            aws_secret_access_key=self.config.DRIVERS['s3']['secret'],
        )

        s3 = session.resource('s3')

        s3.meta.client.upload_file(
            file_location,
            self.config.DRIVERS['s3']['bucket'],
            filename
        )

        return filename



"""A WhiteNoiseProvider Service Provider."""

from whitenoise import WhiteNoise

from masonite.provider import ServiceProvider


class WhitenoiseProvider(ServiceProvider):

    wsgi = False

    def register(self):
        pass

    def boot(self):
        """Wrap the WSGI server in a whitenoise container."""
        self.app.bind('WSGI', WhiteNoise(
            self.app.make('WSGI'), root=self.app.make('Application').STATIC_ROOT))

        for location, alias in self.app.make('Storage').STATICFILES.items():
            self.app.make('WSGI').add_files(location, prefix=alias)



# Pure-python development wsgi server.
# Parts are borrowed from Adrian Holovaty and the Django Project
#  (https://www.djangoproject.com/).

import logging
import socket
import sys
from wsgiref import simple_server

logging.basicConfig(level=logging.INFO)


def is_broken_pipe_error():
    exc_type, exc_value = sys.exc_info()[:2]
    return issubclass(exc_type, socket.error) and exc_value.args[0] == 32


class WSGIServer(simple_server.WSGIServer):
    """BaseHTTPServer that implements the Python WSGI protocol"""

    request_queue_size = 10

    def __init__(self, *args, ipv6=False, allow_reuse_address=True, **kwargs):
        if ipv6:
            self.address_family = socket.AF_INET6
        self.allow_reuse_address = allow_reuse_address
        super().__init__(*args, **kwargs)

    def handle_error(self, request, client_address):
        if is_broken_pipe_error():
            logging.info("- Broken pipe from %s\n", client_address)
        else:
            super().handle_error(request, client_address)


class ServerHandler(simple_server.ServerHandler):
    http_version = '1.1'

    def handle_error(self):
        # Ignore broken pipe errors, otherwise pass on
        if not is_broken_pipe_error():
            super().handle_error()


class WSGIRequestHandler(simple_server.WSGIRequestHandler):
    protocol_version = 'HTTP/1.1'

    def address_string(self):
        # Short-circuit parent method to not call socket.getfqdn
        return self.client_address[0]

    def log_message(self, format, *args):
        extra = {
            'request': self.request,
            'server_time': self.log_date_time_string(),
        }
        if args[1][0] == '4':
            # 0x16 = Handshake, 0x03 = SSL 3.0 or TLS 1.x
            if args[0].startswith('\x16\x03'):
                extra['status_code'] = 500
                logger.error(
                    "You're accessing the development server over HTTPS, but "
                    "it only supports HTTP.\n", extra=extra,
                )
                return

        if args[1].isdigit() and len(args[1]) == 3:
            status_code = int(args[1])
            extra['status_code'] = status_code

            if status_code >= 500:
                level = logging.error
            elif status_code >= 400:
                level = logging.warning
            else:
                level = logging.info
        else:
            level = loggig.info

        level(format, *args, extra=extra)

    def get_environ(self):
        # Strip all headers with underscores in the name before constructing
        # the WSGI environ. This prevents header-spoofing based on ambiguity
        # between underscores and dashes both normalized to underscores in WSGI
        # env vars. Nginx and Apache 2.4+ both do this as well.
        for k in self.headers:
            if '_' in k:
                del self.headers[k]

        return super().get_environ()

    def handle(self):
        """Copy of WSGIRequestHandler.handle() but with different ServerHandler"""
        self.raw_requestline = self.rfile.readline(65537)
        if len(self.raw_requestline) > 65536:
            self.requestline = ''
            self.request_version = ''
            self.command = ''
            self.send_error(414)
            return

        if not self.parse_request():  # An error code has been sent, just exit
            return

        handler = ServerHandler(
            self.rfile, self.wfile, self.get_stderr(), self.get_environ()
        )
        handler.request_handler = self      # backpointer for logging
        handler.run(self.server.get_app())


def _split_module_and_app(moduleapp):
    if ":" in moduleapp:
        parts = moduleapp.split(":")
        return parts[0], parts[1]
    return moduleapp, "application"


def _import_application(module_name, app_name):
    from importlib import import_module
    module = import_module(module_name)
    return getattr(module, app_name)


def run(host, port, wsgi_handler, ipv6=False, httpd_cls=WSGIServer):
    if type(wsgi_handler) == str:
        module_name, app_name = _split_module_and_app(wsgi_handler)
        wsgi_handler = _import_application(module_name, app_name)

    server_address = (host, int(port))
    httpd = httpd_cls(server_address, WSGIRequestHandler, ipv6=ipv6)
    httpd.set_app(wsgi_handler)
    try:
        print('Serving at: http://{}:{}'.format(host, port))
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass



from setuptools import setup
from masonite.info import VERSION

setup(
    name='masonite',
    packages=[
        'masonite',
        'masonite.auth',
        'masonite.providers',
        'masonite.commands',
        'masonite.drivers',
        'masonite.managers',
        'masonite.testsuite',
        'masonite.queues',
        'masonite.contracts',
        'masonite.contracts.managers',
        'masonite.helpers',
        'masonite.middleware',
        'masonite.testing',
    ],
    version=VERSION,
    install_requires=[
        'masonite-events>=1.0,<2',
        'validator.py>=1.2,<1.3',
        'cryptography>=2.3,<2.4',
        'bcrypt>=3.1,<3.2',
        'requests>=2.0,<2.99',
        'tldextract>=2.2,<2.3',
        'Jinja2>=2,<3',
        'python-dotenv>=0.8,<0.9',
        'passlib>=1.7,<1.8',
        'whitenoise>=3.3,<3.4',
        'pytest>=3.6,<4',
        'masonite-entry>=0.0.0,<=0.9.99',
        'masonite-scheduler>=1.0.0,<=1.0.99',
        'pendulum>=1.5,<1.6',
        'cleo>=0.6,<0.7',
        'tabulate>=0.8,<0.9',
        'psutil>=5.4,<5.5',
        'orator>=0.9,<1',
        'hupper>=1.0,<2.0'
    ],
    description='The core for the Masonite framework',
    author='Joseph Mancuso',
    author_email='idmann509@gmail.com',
    url='https://github.com/MasoniteFramework/masonite',
    keywords=['masonite', 'python web framework', 'python3'],
    licence='MIT',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'Operating System :: OS Independent',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',


        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    include_package_data=True,
)



class BaseRequest:
    def user(self, obj):
        self._user = obj
        self.container.on_resolve('Request', self._bind_user_to_request)
        wsgi = generate_wsgi()
        wsgi['PATH_INFO'] = self.url
        self._run_container(wsgi)

        return self

    def ok(self):
        return self.status('200 OK')

    def status(self, value=None):
        if not value:
            return self.container.make('Request').get_status_code()

        return self.container.make('Request').get_status_code() == value



"""Run Migration Command."""
import os
import sys

from subprocess import check_output
from cleo import Command
from masonite.packages import add_venv_site_packages


class MigrateCommand(Command):
    """
    Run migrations.

    migrate
    """

    def handle(self):
        sys.path.append(os.getcwd())
        try:
            add_venv_site_packages()
        except ImportError:
            self.comment(
                'This command must be ran inside of the root of a Masonite project directory')

        from wsgi import container

        migration_directory = ['databases/migrations']
        for key, value in container.providers.items():
            if type(key) == str and 'MigrationDirectory' in key:
                migration_directory.append(value)

        for directory in migration_directory:
            self.line('')
            if len(migration_directory) > 1:
                self.info('Migrating: {}'.format(directory))
            try:
                output = bytes(check_output(
                    ['orator', 'migrate', '-c',
                        'config/database.py', '-p', directory, '-f']
                )).decode('utf-8')

                self.line(
                    output.replace('OK', '<info>OK</info>').replace('Migrated', '<info>Migrated</info><fg=cyan>') + '</>'
                )
            except Exception:
                pass



"""Module for the LoadEnvironment class."""

import os
from pathlib import Path


class LoadEnvironment:
    """This class is used for loading the environment from .env and .env.* files."""

    def __init__(self, env=None, override=True, only=None):
        """LoadEnvironment constructor.

        Keyword Arguments:
            env {string} -- An additional environment file that you want to load. (default: {None})
            override {bool} -- Whether or not the environment variables found should overwrite existing ones. (default: {False})
            only {string} -- If this is set then it will only load that environment. (default: {None})
        """
        from dotenv import load_dotenv
        self.env = load_dotenv

        if only:
            self._load_environment(only, override=override)
            return

        env_path = str(Path('.') / '.env')
        self.env(env_path, override=override)

        if os.environ.get('APP_ENV'):
            self._load_environment(os.environ.get(
                'APP_ENV'), override=override)
        if env:
            self._load_environment(env, override=override)

    def _load_environment(self, env, override=False):
        """Load the environment depending on the env file.

        Arguments:
            env {string} -- Name of the environment file to load from

        Keyword Arguments:
            override {bool} -- Whether the environment file should overwrite existing environment keys. (default: {False})
        """
        env_path = str(Path('.') / '.env.{}'.format(env))
        self.env(dotenv_path=env_path, override=override)


def env(value, default='', cast=True):
    env_var = os.getenv(value, default)

    if not cast:
        return env_var

    if env_var == "":
        env_var = default

    if isinstance(env_var, bool):
        return env_var
    elif env_var is None:
        return None
    elif env_var.isnumeric():
        return int(env_var)
    elif env_var in ('false', 'False'):
        return False
    elif env_var in ('true', 'True'):
        return True
    else:
        return env_var



"""An AppProvider Service Provider."""

from config import application, middleware, storage
from masonite.autoload import Autoload
from masonite.commands import (AuthCommand, CommandCommand, ControllerCommand,
                               DownCommand, InfoCommand, InstallCommand,
                               JobCommand, KeyCommand, MakeMigrationCommand,
                               MiddlewareCommand, MigrateCommand,
                               MigrateRefreshCommand, MigrateResetCommand,
                               MigrateRollbackCommand, MigrateStatusCommand,
                               ModelCommand, ModelDocstringCommand,
                               ProviderCommand, QueueWorkCommand,
                               RoutesCommand, SeedCommand, SeedRunCommand,
                               ServeCommand, TinkerCommand, UpCommand,
                               ValidatorCommand, ViewCommand)
from masonite.exception_handler import DumpHandler, ExceptionHandler
from masonite.helpers.routes import flatten_routes
from masonite.hook import Hook
from masonite.provider import ServiceProvider
from masonite.request import Request
from masonite.response import Response
from masonite.routes import Route
from routes import web


class AppProvider(ServiceProvider):

    def register(self):
        self.app.bind('HookHandler', Hook(self.app))
        self.app.bind('WebRoutes', flatten_routes(web.ROUTES))
        self.app.bind('Storage', storage)
        self.app.bind('Route', Route())
        self.app.bind('Request', Request())
        self.app.simple(Response(self.app))
        self.app.bind('Container', self.app)
        self.app.bind('ExceptionHandler', ExceptionHandler(self.app))
        self.app.bind('ExceptionDumpExceptionHandler', DumpHandler)
        self.app.bind('RouteMiddleware', middleware.ROUTE_MIDDLEWARE)
        self.app.bind('HttpMiddleware', middleware.HTTP_MIDDLEWARE)

        # Insert Commands
        self._load_commands()

        self._autoload(application.AUTOLOAD)

    def boot(self, request: Request, route: Route):
        self.app.bind('StatusCode', '404 Not Found')
        route.load_environ(self.app.make('Environ'))
        request.load_environ(self.app.make('Environ')).load_app(self.app)

    def _autoload(self, directories):
        Autoload(self.app).load(directories)

    def _load_commands(self):
        self.app.bind('MasoniteAuthCommand', AuthCommand())
        self.app.bind('MasoniteCommandCommand', CommandCommand())
        self.app.bind('MasoniteControllerCommand', ControllerCommand())
        self.app.bind('MasoniteDownCommand', DownCommand())
        self.app.bind('MasoniteInfoCommand', InfoCommand())
        self.app.bind('MasoniteInstallCommand', InstallCommand())
        self.app.bind('MasoniteJobCommand', JobCommand())
        self.app.bind('MasoniteKeyCommand', KeyCommand())
        self.app.bind('MasoniteMakeMigrationCommand', MakeMigrationCommand())
        self.app.bind('MasoniteMiddlewareCommand', MiddlewareCommand())
        self.app.bind('MasoniteMigrateCommand', MigrateCommand())
        self.app.bind('MasoniteMigrateRefreshCommand', MigrateRefreshCommand())
        self.app.bind('MasoniteMigrateResetCommand', MigrateResetCommand())
        self.app.bind('MasoniteMigrateStatusCommand', MigrateStatusCommand())
        self.app.bind('MasoniteMigrateRollbackCommand',
                      MigrateRollbackCommand())
        self.app.bind('MasoniteModelCommand', ModelCommand())
        self.app.bind('MasoniteModelDocstringCommand', ModelDocstringCommand())
        self.app.bind('MasoniteProviderCommand', ProviderCommand())
        self.app.bind('MasoniteQueueWorkCommand', QueueWorkCommand())
        self.app.bind('MasoniteViewCommand', ViewCommand())
        self.app.bind('MasoniteRoutesCommand', RoutesCommand())
        self.app.bind('MasoniteServeCommand', ServeCommand())
        self.app.bind('MasoniteSeedCommand', SeedCommand())
        self.app.bind('MasoniteSeedRunCommand', SeedRunCommand())
        self.app.bind('MasoniteTinkerCommand', TinkerCommand())
        self.app.bind('MasoniteUpCommand', UpCommand())
        self.app.bind('MasoniteValidatorCommand', ValidatorCommand())



"""Log Driver Module."""

import logging
import os

from masonite.app import App
from masonite.contracts.MailContract import MailContract
from masonite.drivers.BaseMailDriver import BaseMailDriver
from masonite.view import View


class MailLogDriver(BaseMailDriver, MailContract):
    """Mail log driver.
    """
    def __init__(self, app: App, view: View):
        super().__init__(app, view)

        if 'log' in self.config.DRIVERS and 'location' in self.config.DRIVERS['log']:
            log_location = self.config.DRIVERS['log']['location']
        else:
            log_location = 'bootstrap/mail'

        if not os.path.exists(log_location):
            # Create the path to the model if it does not exist
            os.makedirs(log_location)

        handler = logging.FileHandler('{0}/{1}'.format(
            log_location,
            os.getenv('MAIL_LOGFILE', 'mail.log')
        ))
        self.logger = logging.getLogger(__name__)
        self.logger.handlers = []
        self.logger.propagate = False
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)

    def send(self, message=None):
        """Prints the message in a log.

        Keyword Arguments:
            message {string} -- The message to be printed. (default: { None })

        Returns:
            None
        """

        if not message:
            message = self.message_body

        self.logger.info('***************************************')

        self.logger.info('To: {}'.format(self.to_address))
        self.logger.info('From: {0} <{1}>'.format(
            self.config.FROM['name'], self.config.FROM['address']))
        self.logger.info('Subject: {}'.format(self.message_subject))
        self.logger.info('Message: ')
        self.logger.info(message)

        self.logger.info('***************************************')



"""Session Cookie Module."""

import json

from masonite.contracts import SessionContract
from masonite.drivers import BaseDriver
from masonite.request import Request
from masonite.app import App


class SessionCookieDriver(SessionContract, BaseDriver):
    """Cookie Session Driver."""

    def __init__(self, request: Request, app: App):
        """Cookie Session Constructor.

        Arguments:
            Environ {dict} -- The WSGI environment
            Request {masonite.request.Request} -- The Request class.
        """
        self.environ = app.make('Environ')
        self.request = request

    def get(self, key):
        """Get a value from the session.

        Arguments:
            key {string} -- The key to get from the session.

        Returns:
            string|None - Returns None if a value does not exist.
        """
        cookie = self.request.get_cookie('s_{0}'.format(key))
        if cookie:
            return self._get_serialization_value(cookie)

        return None

    def set(self, key, value):
        """Set a vlue in the session.

        Arguments:
            key {string} -- The key to set as the session key.
            value {string} -- The value to set in the session.
        """
        if isinstance(value, dict):
            value = json.dumps(value)

        self.request.cookie('s_{0}'.format(key), value)

    def has(self, key):
        """Check if a key exists in the session.

        Arguments:
            key {string} -- The key to check for in the session.

        Returns:
            bool
        """
        if self.get(key):
            return True
        return False

    def all(self):
        """Get all session data.

        Returns:
            dict
        """
        return self.__collect_data()

    def delete(self, key):
        """Delete a value in the session by it's key.

        Arguments:
            key {string} -- The key to find in the session.

        Returns:
            bool -- If the key was deleted or not
        """
        self.__collect_data()

        if self.request.get_cookie('s_{}'.format(key)):
            self.request.delete_cookie('s_{}'.format(key))
            return True

        return False

    def __collect_data(self):
        """Collect data from session and flash data.

        Returns:
            dict
        """
        cookies = {}
        if 'HTTP_COOKIE' in self.environ and self.environ['HTTP_COOKIE']:
            cookies_original = self.environ['HTTP_COOKIE'].split(';')
            for cookie in cookies_original:
                if cookie.startswith('s_'):
                    data = cookie.split("=")
                    cookie_value = self.request.get_cookie(data[0])
                    cookies[data[0][2:]] = cookie_value
        return cookies

    def flash(self, key, value):
        """Add temporary data to the session.

        Arguments:
            key {string} -- The key to set as the session key.
            value {string} -- The value to set in the session.
        """
        if isinstance(value, dict):
            value = json.dumps(value)

        self.request.cookie('s_{0}'.format(key), value, expires='2 seconds')

    def reset(self, flash_only=False):
        """Delete all session data.

        Keyword Arguments:
            flash_only {bool} -- If only flash data should be deleted. (default: {False})
        """
        cookies = self.__collect_data()
        for cookie in cookies:
            self.request.delete_cookie('s_{0}'.format(cookie))

    def helper(self):
        """Use to create builtin helper function."""
        return self

    def _get_serialization_value(self, value):
        try:
            return json.loads(value)
        except ValueError:
            return value
